/* ***********************************************************
 * This file was automatically generated on 2019-09-17.      *
 *                                                           *
 * OpenHAB Bindings Version 2.1.23                           *
 *                                                           *
 * If you have a bugfix for this file and want to commit it, *
 * please fix the bug in the generator. You can find a link  *
 * to the generators git repository on tinkerforge.com       *
 *************************************************************/

package com.tinkerforge;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.Arrays;
import java.util.List;
import java.net.URI;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Collections;
import java.util.function.Function;
import java.util.function.BiConsumer;
import org.eclipse.smarthome.config.core.Configuration;
import org.eclipse.smarthome.config.core.ConfigDescription;
import org.eclipse.smarthome.config.core.ConfigDescriptionParameter.Type;
import org.eclipse.smarthome.config.core.ConfigDescriptionParameterBuilder;
import org.eclipse.smarthome.config.core.ConfigDescriptionParameterGroup;
import org.eclipse.smarthome.config.core.ParameterOption;
import org.eclipse.smarthome.core.types.State;
import org.eclipse.smarthome.core.types.StateOption;
import org.eclipse.smarthome.core.types.Command;
import org.eclipse.smarthome.core.types.CommandDescriptionBuilder;
import org.eclipse.smarthome.core.types.CommandOption;
import org.eclipse.smarthome.core.thing.ThingTypeUID;
import org.eclipse.smarthome.core.thing.type.ChannelDefinitionBuilder;
import org.eclipse.smarthome.core.thing.type.ChannelType;
import org.eclipse.smarthome.core.thing.type.ChannelTypeBuilder;
import org.eclipse.smarthome.core.thing.type.ChannelTypeUID;
import org.eclipse.smarthome.core.thing.type.ThingType;
import org.eclipse.smarthome.core.thing.type.ThingTypeBuilder;
import org.eclipse.smarthome.core.types.StateDescriptionFragmentBuilder;
import org.eclipse.smarthome.binding.tinkerforge.internal.TinkerforgeBindingConstants;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.eclipse.smarthome.core.thing.CommonTriggerEvents;
import org.eclipse.smarthome.core.library.types.OnOffType;
import org.eclipse.smarthome.core.library.types.StringType;

/**
 * Controls AC and DC Solid State Relays
 */
public class BrickletSolidStateRelay extends Device {
	public final static int DEVICE_IDENTIFIER = 244;
	public final static String DEVICE_DISPLAY_NAME = "Solid State Relay Bricklet";

    public final static DeviceInfo DEVICE_INFO = new DeviceInfo(DEVICE_DISPLAY_NAME, "solidstaterelay", DEVICE_IDENTIFIER, BrickletSolidStateRelay.class);

	public final static byte FUNCTION_SET_STATE = (byte)1;
	public final static byte FUNCTION_GET_STATE = (byte)2;
	public final static byte FUNCTION_SET_MONOFLOP = (byte)3;
	public final static byte FUNCTION_GET_MONOFLOP = (byte)4;
	public final static byte FUNCTION_GET_IDENTITY = (byte)255;
	private final static int CALLBACK_MONOFLOP_DONE = 5;


	private List<MonoflopDoneListener> listenerMonoflopDone = new CopyOnWriteArrayList<MonoflopDoneListener>();

	public class Monoflop {
		public boolean state;
		public long time;
		public long timeRemaining;

		public String toString() {
			return "[" + "state = " + state + ", " + "time = " + time + ", " + "timeRemaining = " + timeRemaining + "]";
		}
	}

	/**
	 * This listener is triggered whenever the monoflop timer reaches 0.
	 * The parameter is the current state of the relay
	 * (the state after the monoflop).
	 */
	public interface MonoflopDoneListener extends DeviceListener {
		public void monoflopDone(boolean state);
	}

	/**
	 * Creates an object with the unique device ID \c uid. and adds it to
	 * the IP Connection \c ipcon.
	 */
	public BrickletSolidStateRelay(String uid, IPConnection ipcon) {
		super(uid, ipcon);

		apiVersion[0] = 2;
		apiVersion[1] = 0;
		apiVersion[2] = 0;
		responseExpected[IPConnection.unsignedByte(FUNCTION_SET_STATE)] = RESPONSE_EXPECTED_FLAG_FALSE;
		responseExpected[IPConnection.unsignedByte(FUNCTION_GET_STATE)] = RESPONSE_EXPECTED_FLAG_ALWAYS_TRUE;
		responseExpected[IPConnection.unsignedByte(FUNCTION_SET_MONOFLOP)] = RESPONSE_EXPECTED_FLAG_FALSE;
		responseExpected[IPConnection.unsignedByte(FUNCTION_GET_MONOFLOP)] = RESPONSE_EXPECTED_FLAG_ALWAYS_TRUE;
		responseExpected[IPConnection.unsignedByte(FUNCTION_GET_IDENTITY)] = RESPONSE_EXPECTED_FLAG_ALWAYS_TRUE;

		callbacks[CALLBACK_MONOFLOP_DONE] = new IPConnection.DeviceCallbackListener() {
			public void callback(byte[] packet) {
				ByteBuffer bb = ByteBuffer.wrap(packet, 8, packet.length - 8);
				bb.order(ByteOrder.LITTLE_ENDIAN);

				boolean state = (bb.get()) != 0;

				for (MonoflopDoneListener listener: listenerMonoflopDone) {
					listener.monoflopDone(state);
				}
			}
		};
	}

	/**
	 * Sets the state of the relays *true* means on and *false* means off.
	 * 
	 * A running monoflop timer will be aborted if this function is called.
	 * 
	 * The default value is *false*.
	 */
	public void setState(boolean state) throws TinkerforgeException {
		ByteBuffer bb = ipcon.createRequestPacket((byte)9, FUNCTION_SET_STATE, this);

		bb.put((byte)(state ? 1 : 0));

		sendRequest(bb.array());
	}

	/**
	 * Returns the state of the relay, *true* means on and *false* means off.
	 */
	public boolean getState() throws TinkerforgeException {
		ByteBuffer bb = ipcon.createRequestPacket((byte)8, FUNCTION_GET_STATE, this);


		byte[] response = sendRequest(bb.array());

		bb = ByteBuffer.wrap(response, 8, response.length - 8);
		bb.order(ByteOrder.LITTLE_ENDIAN);

		boolean state = (bb.get()) != 0;

		return state;
	}

	/**
	 * The first parameter  is the desired state of the relay (*true* means on
	 * and *false* means off). The second parameter indicates the time (in ms) that
	 * the relay should hold the state.
	 * 
	 * If this function is called with the parameters (true, 1500):
	 * The relay will turn on and in 1.5s it will turn off again.
	 * 
	 * A monoflop can be used as a failsafe mechanism. For example: Lets assume you
	 * have a RS485 bus and a Solid State Relay Bricklet connected to one of the slave
	 * stacks. You can now call this function every second, with a time parameter
	 * of two seconds. The relay will be on all the time. If now the RS485
	 * connection is lost, the relay will turn off in at most two seconds.
	 */
	public void setMonoflop(boolean state, long time) throws TinkerforgeException {
		ByteBuffer bb = ipcon.createRequestPacket((byte)13, FUNCTION_SET_MONOFLOP, this);

		bb.put((byte)(state ? 1 : 0));
		bb.putInt((int)time);

		sendRequest(bb.array());
	}

	/**
	 * Returns the current state and the time as set by
	 * {@link BrickletSolidStateRelay#setMonoflop(boolean, long)} as well as the remaining time until the state flips.
	 * 
	 * If the timer is not running currently, the remaining time will be returned
	 * as 0.
	 */
	public Monoflop getMonoflop() throws TinkerforgeException {
		ByteBuffer bb = ipcon.createRequestPacket((byte)8, FUNCTION_GET_MONOFLOP, this);


		byte[] response = sendRequest(bb.array());

		bb = ByteBuffer.wrap(response, 8, response.length - 8);
		bb.order(ByteOrder.LITTLE_ENDIAN);

		Monoflop obj = new Monoflop();
		obj.state = (bb.get()) != 0;
		obj.time = IPConnection.unsignedInt(bb.getInt());
		obj.timeRemaining = IPConnection.unsignedInt(bb.getInt());

		return obj;
	}

	/**
	 * Returns the UID, the UID where the Bricklet is connected to,
	 * the position, the hardware and firmware version as well as the
	 * device identifier.
	 * 
	 * The position can be 'a', 'b', 'c' or 'd'.
	 * 
	 * The device identifier numbers can be found :ref:`here &lt;device_identifier&gt;`.
	 * |device_identifier_constant|
	 */
	public Identity getIdentity() throws TinkerforgeException {
		ByteBuffer bb = ipcon.createRequestPacket((byte)8, FUNCTION_GET_IDENTITY, this);


		byte[] response = sendRequest(bb.array());

		bb = ByteBuffer.wrap(response, 8, response.length - 8);
		bb.order(ByteOrder.LITTLE_ENDIAN);

		Identity obj = new Identity();
		obj.uid = IPConnection.string(bb, 8);
		obj.connectedUid = IPConnection.string(bb, 8);
		obj.position = (char)(bb.get());
		for (int i = 0; i < 3; i++) {
			obj.hardwareVersion[i] = IPConnection.unsignedByte(bb.get());
		}
		for (int i = 0; i < 3; i++) {
			obj.firmwareVersion[i] = IPConnection.unsignedByte(bb.get());
		}
		obj.deviceIdentifier = IPConnection.unsignedShort(bb.getShort());

		return obj;
	}

	/**
	 * Adds a MonoflopDone listener.
	 */
	public void addMonoflopDoneListener(MonoflopDoneListener listener) {
		listenerMonoflopDone.add(listener);
	}

	/**
	 * Removes a MonoflopDone listener.
	 */
	public void removeMonoflopDoneListener(MonoflopDoneListener listener) {
		listenerMonoflopDone.remove(listener);
	}

    private final Logger logger = LoggerFactory.getLogger(BrickletSolidStateRelay.class);
    private final static Logger static_logger = LoggerFactory.getLogger(BrickletSolidStateRelay.class);

    @Override
    public void initialize(org.eclipse.smarthome.config.core.Configuration config, Function<String, org.eclipse.smarthome.config.core.Configuration> getChannelConfigFn, BiConsumer<String, org.eclipse.smarthome.core.types.State> updateStateFn, BiConsumer<String, String> triggerChannelFn) throws TinkerforgeException {
        BrickletSolidStateRelayConfig cfg = (BrickletSolidStateRelayConfig) config.as(BrickletSolidStateRelayConfig.class);
        this.addMonoflopDoneListener((state) -> {if(true) {updateStateFn.accept("SolidStateRelayRelay", transformSolidStateRelayRelayCallback0(state, cfg));}});
        
		{
		BrickletSolidStateRelayRelayConfig channelCfg = getChannelConfigFn.apply("SolidStateRelayRelay").as(BrickletSolidStateRelayRelayConfig.class);
		
		}
		{
		BrickletSolidStateRelayMonoflopConfig channelCfg = getChannelConfigFn.apply("SolidStateRelayMonofloprelay").as(BrickletSolidStateRelayMonoflopConfig.class);
		
		}
    }

    @Override
    public void dispose(org.eclipse.smarthome.config.core.Configuration config) throws TinkerforgeException {
        BrickletSolidStateRelayConfig cfg = (BrickletSolidStateRelayConfig) config.as(BrickletSolidStateRelayConfig.class);
        this.listenerMonoflopDone.clear();
        
		
    }

    @Override
    public List<String> getEnabledChannels(org.eclipse.smarthome.config.core.Configuration config) throws TinkerforgeException{
        BrickletSolidStateRelayConfig cfg = (BrickletSolidStateRelayConfig) config.as(BrickletSolidStateRelayConfig.class);
        List<String> result = new ArrayList<String>();
        result.add("SolidStateRelayRelay");
		result.add("SolidStateRelayMonofloprelay");
        return result;
    }

    public static ChannelType getChannelType(ChannelTypeUID channelTypeUID) {
        switch(channelTypeUID.getId()) {
            case "SolidStateRelayRelay":
                return ChannelTypeBuilder.state(new ChannelTypeUID("tinkerforge", "SolidStateRelayRelay"), "NOT USED", "Switch").withConfigDescriptionURI(URI.create("channel-type:tinkerforge:SolidStateRelayRelay")).withDescription("NOT USED").build();
            case "SolidStateRelayMonoflop":
                return ChannelTypeBuilder.state(new ChannelTypeUID("tinkerforge", "SolidStateRelayMonoflop"), "NOT USED", "String").withConfigDescriptionURI(URI.create("channel-type:tinkerforge:SolidStateRelayMonoflop")).withDescription("Triggers a monoflop as configured")
.withCommandDescription(CommandDescriptionBuilder.create().withCommandOption(new CommandOption("TRIGGER", "Trigger")).build()).build();
            default:
                static_logger.debug("Unknown channel type ID {}", channelTypeUID.getId());
                break;
        }

        return null;
    }

    public static ThingType getThingType(ThingTypeUID thingTypeUID) {
        return ThingTypeBuilder.instance(thingTypeUID, "Tinkerforge Solid State Relay Bricklet").isListed(false).withSupportedBridgeTypeUIDs(Arrays.asList(TinkerforgeBindingConstants.THING_TYPE_BRICK_DAEMON.getId())).withConfigDescriptionURI(URI.create("thing-type:tinkerforge:" + thingTypeUID.getId())).withDescription("Controls AC and DC Solid State Relays").withChannelDefinitions(Arrays.asList(new ChannelDefinitionBuilder("SolidStateRelayRelay", new ChannelTypeUID("tinkerforge", "SolidStateRelayRelay")).withLabel("Relay").withDescription("Switches the relay. A running monoflop timer for the relay will be aborted if the relay is toggled by this channel.").build(), new ChannelDefinitionBuilder("SolidStateRelayMonofloprelay", new ChannelTypeUID("tinkerforge", "SolidStateRelayMonoflop")).withLabel("Monoflop Relay").build())).build();
    }

    public static ConfigDescription getConfigDescription(URI uri) {
        switch(uri.toASCIIString()) {
            case "thing-type:tinkerforge:solidstaterelay":
                return new ConfigDescription(uri, Arrays.asList(), Arrays.asList(new ConfigDescriptionParameterGroup("update_intervals", "null", true, "Update Intervals", "Update Intervals")));
            case "channel-type:tinkerforge:SolidStateRelayRelay":
                return new ConfigDescription(uri, Arrays.asList(ConfigDescriptionParameterBuilder.create("updateInterval", Type.INTEGER).withDefault("1000").withDescription("Specifies the update interval in milliseconds. A value of 0 disables automatic updates.").withGroupName("update_intervals").withLabel("Update Interval").withUnit("ms").build()));
            case "channel-type:tinkerforge:SolidStateRelayMonoflop":
                return new ConfigDescription(uri, Arrays.asList(ConfigDescriptionParameterBuilder.create("monoflopDuration", Type.INTEGER).withDefault("1000").withDescription("The time (in ms) that the relay should hold the configured value.").withLabel("Monoflop Duration").withUnit("ms").withMinimum(BigDecimal.valueOf(0)).withMaximum(BigDecimal.valueOf(2147483647)).build(), ConfigDescriptionParameterBuilder.create("monoflopValue", Type.BOOLEAN).withDefault("true").withDescription("The desired value of the specified channel. Activated means relay closed and Deactivated means relay open.").withLabel("Monoflop Value").build()));
            default:
                static_logger.debug("Unknown config description URI {}", uri.toASCIIString());
                break;
        }
        return null;
    }

        @Override
    public void refreshValue(String value, org.eclipse.smarthome.config.core.Configuration config, org.eclipse.smarthome.config.core.Configuration channelConfig, BiConsumer<String, org.eclipse.smarthome.core.types.State> updateStateFn, BiConsumer<String, String> triggerChannelFn) throws TinkerforgeException {
        BrickletSolidStateRelayConfig cfg = (BrickletSolidStateRelayConfig) config.as(BrickletSolidStateRelayConfig.class);
        switch(value) {
            case "SolidStateRelayRelay": {
                   BrickletSolidStateRelayRelayConfig channelCfg = channelConfig.as(BrickletSolidStateRelayRelayConfig.class);
                   updateStateFn.accept(value, transformSolidStateRelayRelayGetter0(this.getState(), cfg));
                   return;
               }
            case "SolidStateRelayMonofloprelay": {
                   BrickletSolidStateRelayMonoflopConfig channelCfg = channelConfig.as(BrickletSolidStateRelayMonoflopConfig.class);
                   updateStateFn.accept(value, transformSolidStateRelayMonofloprelayGetter0(this.getMonoflop(), cfg));
                   return;
               }
            default:
                logger.warn("Refresh for unknown channel {}", value);
                break;
        }
    }
    

        @Override
    public List<SetterRefresh> handleCommand(org.eclipse.smarthome.config.core.Configuration config, org.eclipse.smarthome.config.core.Configuration channelConfig, String channel, Command command) throws TinkerforgeException {
        List<SetterRefresh> result = new ArrayList<SetterRefresh>();
        BrickletSolidStateRelayConfig cfg = (BrickletSolidStateRelayConfig) config.as(BrickletSolidStateRelayConfig.class);
        switch(channel) {
            case "SolidStateRelayRelay":
                if (command instanceof OnOffType) {
                    BrickletSolidStateRelayRelayConfig channelCfg = channelConfig.as(BrickletSolidStateRelayRelayConfig.class);
                    OnOffType cmd = (OnOffType) command;
                    this.setState(cmd == OnOffType.ON);
                    
                } else {
                    logger.warn("Command type {} not supported for channel {}. Please use a OnOffType.", command.getClass().getName(), channel);
                }
                break;
            case "SolidStateRelayMonofloprelay":
                if (command instanceof StringType) {
                    BrickletSolidStateRelayMonoflopConfig channelCfg = channelConfig.as(BrickletSolidStateRelayMonoflopConfig.class);
                    StringType cmd = (StringType) command;
                    this.setMonoflop(channelCfg.monoflopValue.booleanValue(), channelCfg.monoflopDuration);
                    result.add(new SetterRefresh("SolidStateRelayRelay", 0));
                } else {
                    logger.warn("Command type {} not supported for channel {}. Please use a StringType.", command.getClass().getName(), channel);
                }
                break;
            default:
                logger.warn("Command for unknown channel {}", channel);
        }
        return result;
    }

        private org.eclipse.smarthome.core.types.State transformSolidStateRelayRelayCallback0(boolean state, BrickletSolidStateRelayConfig cfg) {
        return state ? OnOffType.ON : OnOffType.OFF;
    }
	    private org.eclipse.smarthome.core.types.State transformSolidStateRelayRelayGetter0(boolean value, BrickletSolidStateRelayConfig cfg) {
        return value ? OnOffType.ON : OnOffType.OFF;
    }
	    private org.eclipse.smarthome.core.types.State transformSolidStateRelayMonofloprelayGetter0(Monoflop value, BrickletSolidStateRelayConfig cfg) {
        return value.state ? OnOffType.ON : OnOffType.OFF;
    }
}
